import loopy as lp
import numpy as np
import pymbolic.primitives as prim
from dune.perftool.loopy.temporary import DuneTemporaryVariable
from dune.perftool.loopy.symbolic import substitute
from dune.perftool.options import get_option


def add_vcl_vector(knl, iname_inner, iname_outer):
    read_insn_to_vec_instance = dict()
    write_insn_to_vec_instance = dict()
    read_vecs = set()
    write_vecs = set()

    from loopy.symbolic import WalkMapper
    class FindAlias(WalkMapper):
        def __init__(self, id, expr, vecs, insn_to_vec):
            self.id = id
            self.vecs = vecs
            self.insn_to_vec = insn_to_vec
            self.rec(expr)

        def visit(self, expr, **kwargs):
            if isinstance(expr, prim.Subscript) \
                    and expr.aggregate.name.endswith('alias'):
                vec_name = expr.aggregate.name.replace('alias', 'vec')
                self.vecs.add(vec_name)
                self.insn_to_vec.setdefault(self.id, set())
                self.insn_to_vec[self.id].add((vec_name,expr))
                return False
            else:
                return True
    # find all instances where an alias is read from or written to
    # save the name of the corresponding vector and the alias subscript expression
    for insn in knl.instructions:
        if isinstance(insn, lp.MultiAssignmentBase):
            FindAlias(insn.id, insn.expression, read_vecs, read_insn_to_vec_instance)
            if isinstance(insn, lp.Assignment):
                FindAlias(insn.id, insn.assignee, write_vecs, write_insn_to_vec_instance)

    read_insns = [knl.id_to_insn[id] for id in read_insn_to_vec_instance]
    write_insns = [knl.id_to_insn[id] for id in write_insn_to_vec_instance]

    # add new temporaries for vectors
    new_vec_temporaries = dict()
    for vec in read_vecs|write_vecs:
        new_vec_temporaries[vec] = DuneTemporaryVariable(vec, dtype=np.float64, shape=(4,), managed=True,
                                                         scope=lp.temp_var_scope.PRIVATE, dim_tags=('vec',))

    modified_accum_insn = []
    replace_accum_insn = dict()
    vng = knl.get_var_name_generator()
    idg = knl.get_instruction_id_generator()
    for insn in write_insns:
        if isinstance(insn, lp.Assignment):
            # write accum expr as "r = expr + r"
            expr_without_r = prim.Sum(tuple(e for e in insn.expression.children if not e == insn.assignee))
            if expr_without_r == insn.expression:
                continue
            # finde micro inames
            iname_ix = next((i for i in insn.within_inames if i.startswith('micro') and i.endswith("_x")))
            iname_iy = next((i for i in insn.within_inames if i.startswith('micro') and i.endswith("_y")))

            # erstelle a[iy] und b
            identifier_a = vng('a')
            identifier_b = vng('b')
            new_vec_temporaries[identifier_a] = DuneTemporaryVariable(identifier_a, dtype=np.float64, shape=(2, 4,), managed=True,
                                                                      scope=lp.temp_var_scope.PRIVATE, dim_tags=('c', 'vec',))
            new_vec_temporaries[identifier_b] = DuneTemporaryVariable(identifier_b, dtype=np.float64, shape=(4,), managed=True,
                                                                      scope=lp.temp_var_scope.PRIVATE, dim_tags=('vec',))

            var_a = prim.Subscript(prim.Variable(identifier_a), (prim.Variable(iname_iy), prim.Variable(iname_inner)))
            var_b = prim.Subscript(prim.Variable(identifier_b), (prim.Variable(iname_inner),))

            # init a
            id_init_a = idg('insn_init_'+identifier_a)
            modified_accum_insn.append(lp.Assignment(assignee=substitute(var_a,{iname_iy:prim.Variable(iname_iy+'_head')}),
                                                     expression=0,
                                                     id=id_init_a,
                                                     within_inames=(insn.within_inames-frozenset({iname_ix, iname_iy,
                                                                                                  iname_outer}))
                                                                   |frozenset({iname_iy+'_head'}),
                                                     )
                                       )

            # setze werte für a und b
            expr_b = substitute(expr_without_r, {iname_ix:1})
            expr_a = prim.Sum((substitute(expr_without_r, {iname_ix:0}), var_a))

            id_set_a = idg('insn_'+identifier_a)
            id_set_b = idg('insn_'+identifier_b)
            modified_accum_insn.append(lp.Assignment(assignee=var_b,
                                                     expression=expr_b,
                                                     id=id_set_b,
                                                     depends_on=insn.depends_on,
                                                     within_inames=insn.within_inames-frozenset({iname_ix}),
                                                     )
                                       )
            modified_accum_insn.append(lp.Assignment(assignee=var_a,
                                                     expression=expr_a,
                                                     id=id_set_a,
                                                     depends_on=insn.depends_on|frozenset({id_init_a}),
                                                     within_inames=insn.within_inames-frozenset({iname_ix}),
                                                     )
                                       )

            # r+=a[iy]
            id_accum = idg('insn_mod_accum')
            expr_accum = prim.Sum((var_a, prim.Call(prim.Variable('permute4d<-1,0,1,2>'), (var_b,)),
                                   substitute(insn.assignee, {iname_ix:0})))
            replace_accum_insn[insn.id] = lp.Assignment(assignee=substitute(insn.assignee,{iname_ix:0}),
                                                        expression=expr_accum,
                                                        id=id_accum,
                                                        depends_on=insn.depends_on|frozenset({id_set_a,id_init_a,id_set_b}),
                                                        within_inames=insn.within_inames-frozenset({iname_ix})
                                                        )
            # a[iy] = permute
            id_permute = idg('insn_permute')
            expr_permute = prim.Call(prim.Variable('permute4d<3,-1,-1,-1>'), (var_b,))
            modified_accum_insn.append(lp.Assignment(assignee=var_a,
                                                     expression=expr_permute,
                                                     id=id_permute,
                                                     depends_on=replace_accum_insn[insn.id].depends_on
                                                                |frozenset({replace_accum_insn[insn.id].id}),
                                                     within_inames=insn.within_inames-frozenset({iname_ix})
                                                     )
                                       )

            # tail handling               | frozenset({iname_iy+'_tail'})))
            id_accum_tail = idg('insn_accum_tail')
            subst_map = {iname_inner: 0, iname_outer: get_option("number_of_blocks")/4, iname_iy: prim.Variable(iname_iy+'_tail'),
                         iname_ix: 0}
            assignee_tail = substitute(insn.assignee, subst_map)
            expr_tail = prim.Sum((prim.Subscript(prim.Variable(identifier_a), (prim.Variable(iname_iy+'_tail'), 0)),
                                  assignee_tail))
            modified_accum_insn.append(lp.Assignment(assignee=assignee_tail,
                                                     expression=expr_tail,
                                                     id=id_accum_tail,
                                                     depends_on=frozenset({replace_accum_insn[insn.id].id,
                                                                           id_permute, id_set_a, id_init_a}),
                                                     within_inames=(insn.within_inames - frozenset({iname_inner, iname_outer,
                                                                                                    iname_ix, iname_iy}))
                                                                   | frozenset({iname_iy+'_tail'})))

    # update found instances with respect to modified accumulation instruction
    def update_insn_to_vec_instance(insns, expr, insn_to_vec_instance):
        new_insns = []
        for insn in insns:
            if insn.id in replace_accum_insn:
                new_insns.append(replace_accum_insn[insn.id])
                insn_to_vec_instance.pop(insn.id)
                FindAlias(new_insns[-1].id, expr(new_insns[-1]), set(), insn_to_vec_instance)
            else:
                new_insns.append(insn)
        return new_insns
    read_insns = update_insn_to_vec_instance(read_insns, lambda i: i.expression, read_insn_to_vec_instance)
    write_insns = update_insn_to_vec_instance(write_insns, lambda i: i.assignee, write_insn_to_vec_instance)

    # add load instructions if the vector is read, based on the read instruction
    # TODO brauche mehrere vectoren, falls in einer insn von einem alias mit unterschiedlichen idx gelesen wird
    # muss dafür eigentlich nur den namen des vectors anpassen
    load_insns = []
    read_dependencies = dict()
    for insn in read_insns:
        for vec, expr in read_insn_to_vec_instance[insn.id]:
            alias = vec.replace('vec', 'alias')

            # flat index without vec iname
            strides = tuple(tag.stride for tag in knl.arg_dict[alias].dim_tags)
            index = prim.Sum(tuple(prim.Product(z) for z in zip(substitute(expr, {iname_inner:0}).index_tuple, strides)))

            # add load instruction
            load_id = idg('insn_'+vec+'_load')
            call_load = prim.Call(prim.Variable(vec+'.load'), (prim.Sum((prim.Variable(alias), index)),))
            load_insns.append(lp.CallInstruction(assignees=(),expression=call_load,
                                                 id=load_id,within_inames=insn.within_inames|insn.reduction_inames(),))
            read_dependencies.setdefault(insn.id, set())
            read_dependencies[insn.id].add(load_id)

    # add store instructions if the vector is written, based on the write instruction
    store_insns = []
    for insn in write_insns:
        for vec, expr in write_insn_to_vec_instance[insn.id]:
            alias = vec.replace('vec', 'alias')

            # flat index without vec iname
            strides = tuple(tag.stride for tag in knl.arg_dict[alias].dim_tags)
            index = prim.Sum(tuple(prim.Product(z) for z in zip(substitute(expr, {iname_inner: 0, iname_ix: 0}).index_tuple, strides)))

            # add store instruction
            store_id = idg('insn_'+vec+'_store')
            call_store = prim.Call(prim.Variable(vec+'.store'), (prim.Sum((prim.Variable(alias), index)),))
            store_insns.append(lp.CallInstruction(assignees=(),expression=call_store,
                                                  id=store_id,within_inames=insn.within_inames,
                                                  depends_on=insn.depends_on
                                                             | frozenset({insn.id})
                                                             | read_dependencies[insn.id],
                                                  ))

    # exchange alias for vector
    new_insns = []
    for insn in knl.instructions:
        insn = replace_accum_insn.get(insn.id, insn)
        if insn.id not in read_insn_to_vec_instance.keys() | write_insn_to_vec_instance.keys():
            new_insns.append(insn)
        else:
            subst_map = dict()
            for vec, expr in read_insn_to_vec_instance[insn.id]:
                subst_map[expr] = prim.Subscript(prim.Variable(vec), (prim.Variable(iname_inner),))

            new_insn = insn

            if insn in read_insns:
                new_insn = new_insn.copy(expression=substitute(new_insn.expression, subst_map),
                                         depends_on=new_insn.depends_on | read_dependencies[insn.id])
            if insn in write_insns:
                new_insn = new_insn.copy(assignee=substitute(new_insn.assignee, subst_map))

            new_insns.append(new_insn)

    from loopy.kernel.creation import resolve_dependencies
    return resolve_dependencies(knl.copy(instructions=new_insns+load_insns+store_insns+modified_accum_insn,
                                         temporary_variables=dict(**knl.temporary_variables, **new_vec_temporaries)))


def find_accumulation_inames(knl):
    inames = set()
    for insn in knl.instructions:
        if any((n.startswith('r_') and n.endswith('alias') for n in insn.write_dependency_names())):
            inames |= insn.within_inames

    inames = set((i for i in inames if i.startswith('micro') and not i.endswith('_x')))

    return inames


def vectorize_micro_elements(knl):
    if "subel_x" in knl.all_inames():
        # need inames for head and tail handling a priori
        from loopy.match import Not, All
        accum_inames = find_accumulation_inames(knl)
        for iname in accum_inames:
            knl = lp.duplicate_inames(knl, iname, Not(All()), suffix='_tail')
            knl = lp.duplicate_inames(knl, iname, Not(All()), suffix='_head')
        knl = lp.split_iname(knl,"subel_x",4, inner_tag='vec')
        knl = add_vcl_vector(knl,'subel_x_inner', 'subel_x_outer')
    return knl
